"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const t = require("babel-types");
const utils_1 = require("./utils");
const jsx_1 = require("./jsx");
const constant_1 = require("./constant");
const create_html_element_1 = require("./create-html-element");
class Transformer {
    constructor(path, isRoot, componentSourceMap, location) {
        this.result = {
            template: '',
            components: []
        };
        this.methods = new Map();
        this.initState = new Set();
        this.importedJSXElement = new Map();
        this.renderMethod = null;
        this.duplicateComponents = new Map();
        this.classPath = path;
        this.isRoot = isRoot;
        this.componentSourceMap = componentSourceMap;
        this.location = location;
        this.moduleNames = Object.keys(path.scope.getAllBindings('module'));
        this.compile();
    }
    traverse() {
        const self = this;
        self.classPath.traverse({
            ClassMethod(path) {
                const node = path.node;
                if (t.isIdentifier(node.key)) {
                    const name = node.key.name;
                    self.methods.set(name, path);
                    if (name === 'render') {
                        self.renderMethod = path;
                    }
                    if (name === 'constructor') {
                        path.traverse({
                            AssignmentExpression(p) {
                                if (t.isMemberExpression(p.node.left) &&
                                    t.isThisExpression(p.node.left.object) &&
                                    t.isIdentifier(p.node.left.property) &&
                                    p.node.left.property.name === 'state' &&
                                    t.isObjectExpression(p.node.right)) {
                                    const properties = p.node.right.properties;
                                    properties.forEach(p => {
                                        if (t.isObjectProperty(p)) {
                                            if (t.isIdentifier(p.key)) {
                                                self.initState.add(p.key.name);
                                            }
                                        }
                                    });
                                }
                            }
                        });
                    }
                }
            },
            ClassProperty(path) {
                const { key: { name }, value } = path.node;
                if (t.isArrowFunctionExpression(value) || t.isFunctionExpression(value)) {
                    self.methods.set(name, path);
                }
                if (name === 'state' && t.isObjectExpression(value)) {
                    value.properties.forEach(p => {
                        if (t.isObjectProperty(p)) {
                            if (t.isIdentifier(p.key)) {
                                self.initState.add(p.key.name);
                            }
                        }
                    });
                }
            },
            JSXElement(path) {
                const id = path.node.openingElement.name;
                if (t.isJSXIdentifier(id) &&
                    !constant_1.Component_SET.has(id.name) &&
                    self.moduleNames.indexOf(id.name) !== -1) {
                    self.importedJSXElement.set(path, id.name);
                }
            }
        });
    }
    renameImportJSXElement() {
        this.importedJSXElement.forEach((name, path) => {
            if (this.duplicateComponents.has(name)) {
                const times = this.duplicateComponents.get(name);
                this.duplicateComponents.set(name, times + 1);
                this.importedJSXElement.set(path, `${name}_${times + 1}`);
            }
            else {
                this.duplicateComponents.set(name, 0);
            }
        });
    }
    setProps() {
        this.classPath.node.body.body = [
            build$PropsProperty(this.importedJSXElement),
            build$ComponentsProperty(this.importedJSXElement),
            ...this.classPath.node.body.body
        ];
    }
    setComponentResult() {
        this.componentSourceMap.forEach((names, source) => {
            let name = '';
            const isUsedJSX = names.some(n => {
                const bool = this.duplicateComponents.has(n);
                if (bool) {
                    name = n;
                }
                return bool;
            });
            if (isUsedJSX) {
                this.result.components.push({
                    name,
                    path: source
                });
            }
        });
    }
    replaceImportedJSXElement() {
        this.importedJSXElement.forEach((name, path) => {
            path.replaceWith(jsx_1.buildRefTemplate(findImportedName(name), name));
        });
        this.result.template = this.result.components.reduce((acc, { name, path }, index) => {
            return acc + `${index === 0 ? '' : `\n`}` + create_html_element_1.createHTMLElement({
                name: 'import',
                attributes: {
                    src: utils_1.pathResolver(path, this.location)
                },
                value: ''
            });
        }, '');
    }
    resetConstructor() {
        if (this.methods.has('constructor')) {
            const ctor = this.methods.get('constructor');
            if (ctor && ctor.isClassMethod()) {
                ctor.node.body.body.push(resetThisState());
            }
        }
        else {
            const ctor = buildConstructor();
            ctor.body.body.push(resetThisState());
            this.classPath.node.body.body.unshift(ctor);
        }
    }
    compile() {
        this.traverse();
        this.renameImportJSXElement();
        this.setProps();
        this.setComponentResult();
        this.replaceImportedJSXElement();
        this.resetConstructor();
    }
}
function resetThisState() {
    return t.expressionStatement(t.assignmentExpression('=', t.memberExpression(t.thisExpression(), t.identifier('state')), t.callExpression(t.memberExpression(t.thisExpression(), t.identifier('_createData')), [])));
}
function buildConstructor() {
    const ctor = t.classMethod('constructor', t.identifier('constructor'), [], t.blockStatement([
        t.expressionStatement(t.callExpression(t.identifier('super'), [
            t.spreadElement(t.identifier('arguments'))
        ]))
    ]));
    return ctor;
}
function build$PropsProperty(importedJSXElement) {
    const properties = [];
    importedJSXElement.forEach((name, path) => {
        const attrObj = [
            t.objectProperty(t.identifier('$name'), t.stringLiteral(name)),
            ...path.node.openingElement.attributes.map(attr => {
                const name = attr.name;
                const value = attr.value;
                let _value;
                // @TODO: 使用作用域的方法全部替换，现在 cover 的情况太少
                if (t.isJSXElement(value)) {
                    throw utils_1.codeFrameError(attr.loc, 'JSX 参数不支持传入 JSX 元素');
                }
                else if (t.isJSXExpressionContainer(value)) {
                    _value = value.expression;
                    if (t.isIdentifier(_value)) {
                        _value = t.memberExpression(t.memberExpression(t.thisExpression(), t.identifier('state')), _value);
                    }
                }
                else {
                    _value = value;
                }
                return t.objectProperty(t.identifier(name.name), _value);
            })
        ];
        properties.push(t.objectMethod('method', t.identifier(name), [], t.blockStatement([t.returnStatement(t.objectExpression(attrObj))])));
    });
    return t.classProperty(t.identifier('$props'), t.objectExpression(properties));
}
function findImportedName(name) {
    return utils_1.isNumeric(name.slice(-1)) && name.slice(-2)[0] === '_'
        ? name.slice(0, name.length - 2)
        : name;
}
function build$ComponentsProperty(importedJSXElement) {
    const properties = [];
    for (const name of importedJSXElement.values()) {
        properties.push(t.objectProperty(t.identifier(name), t.identifier(findImportedName(name))));
    }
    return t.classProperty(t.identifier('$components'), t.objectExpression(properties));
}
//# sourceMappingURL=_class.js.map